// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2009-2013 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

//#if !DEBUG
#define TRANSPARENT_ANCHOR
//#endif

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Drawing.Text;
using System.IO;
using System.Threading;
using System.Windows.Forms;

using Smdn.Mathematics;
using Smdn.Formats.Thbgm;

#if DEBUG
using System.Diagnostics;
#endif

namespace Smdn.Windows.Forms.ThbgmPlayer {
  public class WaveFormView : Control {
    private class WaveForm : IDisposable {
      public const int WaveFormPartLengthInMilliseconds = 100; // 100ms

      public event EventHandler ReadWaveFormFinished;

      public WaveForm(EffectAppliedBgmStream stream)
      {
        this.stream = stream;
        this.readSampleCount = 0;
        this.leftSamples = EmptyArray<float>.Instance;
        this.rightSamples = EmptyArray<float>.Instance;

        ReadWaveForm();
      }

      public void Dispose()
      {
        ReadWaveFormFinished = null;

        abortReadWaveForm = true;

        lock (streamLockObject) {
          if (stream != null) {
            stream.Close();
            stream = null;
          }
        }

        readSampleCount = 0;

        leftSamples = null;
        rightSamples = null;
      }

      private void ReadWaveForm()
      {
        if (stream == null)
          return;

        abortReadWaveForm = false;
        readWaveFormFinished = false;

        ThreadPool.QueueUserWorkItem((WaitCallback)ReadWaveFormCore);
      }

      private void ReadWaveFormCore(object state)
      {
#if DEBUG
        try {
          var sw = new Stopwatch();
          sw.Reset();
          sw.Start();
#endif
        BinaryReader reader;
        int samplesToRead;

        lock (streamLockObject) {
          if (abortReadWaveForm) {
            abortReadWaveForm = false;
            return;
          }

          reader = new BinaryReader(stream);
          samplesToRead = (stream.Format.SamplesPerSecond * WaveFormPartLengthInMilliseconds) / 1000;
        }

        var totalSampleCount = (int)((stream.Length / stream.Format.BlockAlign) / samplesToRead) + 1;

        if (leftSamples.Length < totalSampleCount)
          Array.Resize(ref leftSamples, totalSampleCount);
        if (rightSamples.Length < totalSampleCount)
          Array.Resize(ref rightSamples, totalSampleCount);

        var divider = 1.0f / Int16.MaxValue;

        for (;;) {
          if (abortReadWaveForm) {
            abortReadWaveForm = false;
            return;
          }

          Int16 left  = 0;
          Int16 right = 0;

          lock (streamLockObject) {
            try {
              left  = reader.ReadInt16();
              right = reader.ReadInt16();
            }
            catch (EndOfStreamException) {
              break;
            }

            reader.BaseStream.Seek((samplesToRead - 1) * stream.Format.BlockAlign, SeekOrigin.Current);
          }

          leftSamples [readSampleCount] = Math.Abs((int)left)  * divider;
          rightSamples[readSampleCount] = Math.Abs((int)right) * divider;
          readSampleCount++;
        }

        readWaveFormFinished = true;

        var ev = this.ReadWaveFormFinished;

        if (ev != null)
          ev(this, EventArgs.Empty);
#if DEBUG
          sw.Stop();
          Console.WriteLine("elapsed {0}, stream impl {1}", sw.Elapsed, stream.InnerStream.GetType());
        }
        catch (Exception ex) {
          Console.WriteLine(ex);
        }
#endif
      }

      public void Get(TimeSpan start, TimeSpan length, out float[] left, out float[] right)
      {
        var startSample  = (int)(start.TotalMilliseconds  / WaveFormPartLengthInMilliseconds);
        var lengthSample = (int)(length.TotalMilliseconds / WaveFormPartLengthInMilliseconds);

        left  = new float[lengthSample];
        right = new float[lengthSample];

        if (readSampleCount < startSample)
          return;

        var bytesToCopy = Math.Min(readSampleCount, lengthSample) * sizeof(float);
        var offset = startSample * sizeof(float);

        Buffer.BlockCopy(leftSamples,  0, left,  offset, bytesToCopy);
        Buffer.BlockCopy(rightSamples, 0, right, offset, bytesToCopy);
      }

      public void Read(TimeSpan start, TimeSpan length, Action<float[], int, int> readLeftWaveForm, Action<float[], int, int> readRightWaveForm)
      {
        var startSample  = (int)(start.TotalMilliseconds  / WaveFormPartLengthInMilliseconds);
        var lengthSample = (int)(length.TotalMilliseconds / WaveFormPartLengthInMilliseconds);

        if (readSampleCount < startSample) {
          readLeftWaveForm (EmptyArray<float>.Instance, 0, 0);
          readRightWaveForm(EmptyArray<float>.Instance, 0, 0);
        }
        else {
          var samplesToRead = Math.Min(readSampleCount, lengthSample);

          readLeftWaveForm (leftSamples, startSample, samplesToRead);
          readRightWaveForm(rightSamples, startSample, samplesToRead);
        }
      }

      public void Discard(TimeSpan start)
      {
        var discardStartSample = (int)(start.TotalMilliseconds  / WaveFormPartLengthInMilliseconds);

        if (discardStartSample < readSampleCount)
          readSampleCount = discardStartSample;

        lock (streamLockObject) {
          if (stream != null)
            stream.Position = stream.Format.ToAlignedByteCount(start);
        }

        if (readWaveFormFinished)
          ReadWaveForm();
      }

      private EffectAppliedBgmStream stream;
      private object streamLockObject = new object();
      private float[] leftSamples;
      private float[] rightSamples;
      private int readSampleCount;
      private bool abortReadWaveForm = false;
      private bool readWaveFormFinished = false;
    }

    private class EffectAnchor : Control {
#if TRANSPARENT_ANCHOR
      protected override CreateParams CreateParams
      {
        get
        {
          const int WS_EX_TRANSPARENT = 0x20;
          var cp = base.CreateParams;
          cp.ExStyle |= WS_EX_TRANSPARENT;
          return cp;
        }
      }
#endif

      public EffectAnchor()
      {
#if TRANSPARENT_ANCHOR
        SetStyle(ControlStyles.Opaque, true);
        UpdateStyles();
#endif
        this.Size = new Size(9, 9);
        this.Cursor = Cursors.SizeWE;
      }

#if !TRANSPARENT_ANCHOR
      protected override void OnPaintBackground(System.Windows.Forms.PaintEventArgs pevent)
      {
        ControlPaint.DrawBorder3D(pevent.Graphics, this.Bounds);

        base.OnPaintBackground(pevent);
      }
#endif
    }

    public event EventHandler<SeekBarMovedEventArgs> SeekBarMoved;

    public string BgmSourcePath {
      get; set;
    }

    public Track Track {
      get { return track; }
      set
      {
        if (track == value)
          return;

        if (track != null)
          track.Effect.PropertyChanged -= Effect_PropertyChanged;

        track = value;

        if (track != null) {
          lastFadingStart = track.Effect.FadingStart;

          track.Effect.PropertyChanged += Effect_PropertyChanged;
        }

        OnTrackChanged();
      }
    }

    public RectangleF XAxisRulerBounds {
      get; private set;
    }

    public RectangleF YAxisRulerBounds {
      get; private set;
    }

    public RectangleF WaveFormBounds {
      get; private set;
    }

    public TimeSpan SeekBarPosition {
      get { return seekBarPosition; }
      set
      {
        if (value < TimeSpan.Zero)
          throw new ArgumentOutOfRangeException("SeekBarPosition", "must be zero or positive value");
        var last = seekBarPosition;
        seekBarPosition = value;
        UpdateSeekBarPosition(last);
      }
    }

    private const float PixelsPerSecondScale = 1.5f;

    private float PixelsPerSecondOffset {
      get { return -hscrollOffset.Value * PixelsPerSecondScale; }
    }

    private float WaveFormBoundsMiddle {
      get; set;
    }

    public WaveFormView()
    {
      this.SetStyle(ControlStyles.DoubleBuffer, true);
      this.SetStyle(ControlStyles.UserPaint, true);
      this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
      
      base.ResizeRedraw = true;

      this.hscrollOffset = new HScrollBar();
      this.hscrollOffset.Dock = DockStyle.Bottom;
      this.hscrollOffset.Height = 14;
      this.hscrollOffset.Minimum = 0;
      this.hscrollOffset.SmallChange = 10; // secs
      this.hscrollOffset.LargeChange = 60; // secs
      this.hscrollOffset.ValueChanged += delegate {
        UpdateAnchorLocation();

        Refresh();
      };

      this.Controls.Add(hscrollOffset);

      this.fadingStartAnchor = new EffectAnchor();
      this.fadingEndAnchor = new EffectAnchor();
      this.trailingSilenceEndAnchor = new EffectAnchor();

      anchors = new[] {fadingStartAnchor, fadingEndAnchor, trailingSilenceEndAnchor};

      foreach (var anchor in anchors) {
        anchor.MouseDown += EffectAnchor_MouseDown;
        anchor.MouseMove += EffectAnchor_MouseMove;
        anchor.MouseUp   += EffectAnchor_MouseUp;
      }

      this.Controls.Add(fadingStartAnchor);
      this.Controls.Add(fadingEndAnchor);
      this.Controls.Add(trailingSilenceEndAnchor);

      foreach (var anchor in anchors) {
        anchor.BringToFront();
      }

      this.fadingStartAnchor.MouseHover += delegate(object sender, EventArgs e) {
        var a = sender as EffectAnchor;
        this.toolTip.Show(Texts.GetText("FadingStartPoint"), a, a.Width, a.Height, 750);
      };
      this.fadingEndAnchor.MouseHover += delegate(object sender, EventArgs e) {
        var a = sender as EffectAnchor;
        this.toolTip.Show(Texts.GetText("FadingEndPoint"), a, a.Width, a.Height, 750);
      };
      this.trailingSilenceEndAnchor.MouseHover += delegate(object sender, EventArgs e) {
        var a = sender as EffectAnchor;
        this.toolTip.Show(Texts.GetText("TrackEndPoint"), a, a.Width, a.Height, 750);
      };

      this.toolTip = new ToolTip();
      this.toolTip.ShowAlways = false;

      components.Add(hscrollOffset);
      components.Add(fadingStartAnchor);
      components.Add(fadingEndAnchor);
      components.Add(trailingSilenceEndAnchor);
      components.Add(toolTip);
    }

    protected override void OnSizeChanged(EventArgs e)
    {
      const float rulerSize = 16.0f;

      XAxisRulerBounds = RectangleF.FromLTRB(rulerSize, 0.0f, (float)Width, rulerSize);
      YAxisRulerBounds = RectangleF.FromLTRB(0.0f, rulerSize, rulerSize, (float)(Height - hscrollOffset.Height));
      WaveFormBounds = RectangleF.FromLTRB(rulerSize, rulerSize, XAxisRulerBounds.Right, YAxisRulerBounds.Bottom);
      WaveFormBoundsMiddle = WaveFormBounds.Top + WaveFormBounds.Height * 0.5f;

      UpdateAnchorLocation();

      DiscardCachedWaveFormImage();

      base.OnSizeChanged(e);
    }

    protected override void Dispose(bool disposing)
    {
      if (disposing) {
        if (components != null)
          components.Dispose();

        if (waveForm != null) {
          waveForm.Dispose();
          waveForm = null;
        }

        DiscardCachedWaveFormImage();
      }

      base.Dispose(disposing);
    }

    protected override void OnCreateControl()
    {
      UpdateScrollRange();

      UpdateAnchorLocation();

      base.OnCreateControl();
    }

    private void OnTrackChanged()
    {
      if (waveForm != null) {
        waveForm.Dispose();
        waveForm = null;
      }

      EffectAppliedBgmStream bgmStream = null;

      try {
        if (track == null || string.IsNullOrEmpty(BgmSourcePath))
          bgmStream = null;
        else
          bgmStream = track.GetStream(BgmSourcePath, true, track.TrackInfo.StreamFormat.BlockAlign);
      }
      catch {
        // ignore exceptions
      }

      waveForm = new WaveForm(bgmStream);
      waveForm.ReadWaveFormFinished += WaveForm_ReadWaveFormFinished;

      DiscardCachedWaveFormImage();

      UpdateView();
    }

    public void UpdateView()
    {
      UpdateScrollRange();

      UpdateAnchorLocation();

      Refresh();
    }

    private void UpdateScrollRange()
    {
      const double padding = 10.0;

      var max = ((track == null) ? 0.0 : track.Effect.TotalLength.TotalSeconds) - (0.25f * WaveFormBounds.Width / PixelsPerSecondScale) + hscrollOffset.LargeChange + padding;

      if (max < hscrollOffset.LargeChange) {
        hscrollOffset.Value = 0;
        hscrollOffset.Enabled = false;
      }
      else {
        hscrollOffset.Enabled = true;
        hscrollOffset.Maximum = (int)max;
      }
    }

    private void UpdateAnchorLocation()
    {
      if (track == null) {
        foreach (var anchor in anchors) {
          anchor.Visible = false;
        }
        return;
      }

      foreach (var pair in new[] {
        new {Anchor = fadingStartAnchor,        Location = track.Effect.FadingStart.TotalSeconds},
        new {Anchor = fadingEndAnchor,          Location = (track.Effect.FadingStart + track.Effect.FadingLength).TotalSeconds},
        new {Anchor = trailingSilenceEndAnchor, Location = track.Effect.TotalLength.TotalSeconds},
      }) {
        pair.Anchor.Location = new Point((int)(pair.Location * PixelsPerSecondScale + PixelsPerSecondOffset + WaveFormBounds.Left - pair.Anchor.Width / 2), (int)WaveFormBoundsMiddle - pair.Anchor.Height / 2);
        pair.Anchor.Visible = YAxisRulerBounds.Right <= pair.Anchor.Right;
      }
    }

    private void UpdateSeekBarPosition(TimeSpan lastSeekBarPosition)
    {
      var newX = (float)(seekBarPosition.TotalSeconds * PixelsPerSecondScale) + PixelsPerSecondOffset + WaveFormBounds.Left;
      var lastX = (float)(lastSeekBarPosition.TotalSeconds * PixelsPerSecondScale) + PixelsPerSecondOffset + WaveFormBounds.Left;

      Invalidate(Rectangle.Ceiling(RectangleF.FromLTRB(newX - 2.0f, WaveFormBounds.Top, newX + 2.0f, WaveFormBounds.Bottom)));
      Invalidate(Rectangle.Ceiling(RectangleF.FromLTRB(lastX - 2.0f, WaveFormBounds.Top, lastX + 2.0f, WaveFormBounds.Bottom)));
      Update();
    }

    protected override void OnPaint(PaintEventArgs e)
    {
      var g = e.Graphics;

      g.Clear(SystemColors.AppWorkspace);
      g.InterpolationMode = InterpolationMode.HighQualityBicubic;
      g.SmoothingMode = SmoothingMode.AntiAlias;

      // ruler
      for (var minute = 0; ; minute++) {
        var x = (minute * 60) * PixelsPerSecondScale + PixelsPerSecondOffset;

        if (x < 0)
          continue;

        if (XAxisRulerBounds.Width < x || XAxisRulerBounds.Width <= 0)
          break;

        x += XAxisRulerBounds.Left;

        g.DrawLine(SystemPens.ControlText, x, XAxisRulerBounds.Bottom - 4.0f, x, XAxisRulerBounds.Bottom);
        g.DrawString(string.Format("{0:N1}min", minute), Control.DefaultFont, SystemBrushes.ControlText, x, XAxisRulerBounds.Bottom - 16.0f);
      }

      // wave form bound
      g.FillRectangle(Brushes.Black, WaveFormBounds);

      // effect and wave form
      if (track == null)
        goto DrawEffectDone;

      var effect = track.Effect;
      var fadeStartX  = (float)(effect.FadingStart.TotalSeconds * PixelsPerSecondScale) + PixelsPerSecondOffset;
      var fadeEndX    = (float)((effect.FadingStart + effect.FadingLength).TotalSeconds * PixelsPerSecondScale) + PixelsPerSecondOffset;
      var silenceEndX = (float)((effect.FadingStart + effect.FadingLength + effect.TrailingSilenceLength).TotalSeconds * PixelsPerSecondScale) + PixelsPerSecondOffset;

      using (var brushFadeEffect = new SolidBrush(Color.FromArgb(0x40, 0x00, 0x40, 0xff))) {
        if (WaveFormBounds.Width < fadeStartX)
          g.FillRectangle(brushFadeEffect, WaveFormBounds);
        else
          g.FillRectangle(brushFadeEffect, WaveFormBounds.Left, WaveFormBounds.Top - 1, fadeStartX, WaveFormBounds.Height);

        g.SmoothingMode = SmoothingMode.None;

        using (var b = new SolidBrush(Color.FromArgb(0x20, 0x20, 0x20))) {
          g.FillRectangle(b, RectangleF.FromLTRB(WaveFormBounds.Left + silenceEndX, WaveFormBounds.Top, WaveFormBounds.Right, WaveFormBounds.Bottom));
        }

        g.DrawLine(SystemPens.ControlText, WaveFormBounds.Left, WaveFormBounds.Top, WaveFormBounds.Right, WaveFormBounds.Top);
        g.DrawLine(SystemPens.ControlText, WaveFormBounds.Left, WaveFormBounds.Top, WaveFormBounds.Left, WaveFormBounds.Bottom);

        g.SmoothingMode = SmoothingMode.AntiAlias;

        if ((fadeStartX < 0.0f && fadeEndX < 0.0f) || (WaveFormBounds.Width < fadeStartX))
          goto DrawFadeEffectDone;

        // fading curve
        PointF[] fadingCurvePoints;

        switch (effect.FadingCurveType) {
          case FadingCurveType.None:
            fadingCurvePoints = new[] {
              new PointF(fadeStartX, 1.0f),
              new PointF(fadeEndX,   1.0f),
            };
            break;

          case FadingCurveType.Linear:
            fadingCurvePoints = new[] {
              new PointF(fadeStartX, 1.0f),
              new PointF(fadeEndX,   0.0f),
            };
            break;

          case FadingCurveType.Exponential:
          case FadingCurveType.Logarithmic:
          case FadingCurveType.Trigonometrical:
            fadingCurvePoints = new PointF[30];

            for (var i = 0; i < fadingCurvePoints.Length; i++) {
              var ratio = i / (double)(fadingCurvePoints.Length - 1);
              var vol = 1.0 - ratio;

              if (effect.FadingCurveType == FadingCurveType.Exponential)
                vol = Effect.CalculateVolumeExponential(vol, effect.FadingCurveCoefficient);
              else if (effect.FadingCurveType == FadingCurveType.Logarithmic)
                vol = Effect.CalculateVolumeLogarithmic(vol, effect.FadingCurveCoefficient);
              else if (effect.FadingCurveType == FadingCurveType.Trigonometrical)
                vol = Effect.CalculateVolumeTrigonometical(vol, effect.FadingCurveCoefficient);

              fadingCurvePoints[i] = new PointF((float)(fadeStartX + (fadeEndX - fadeStartX) * ratio), (float)vol);
            }
            break;

          default:
            fadingCurvePoints = EmptyArray<PointF>.Instance;
            break;
        }

        // draw outline
        var fadingUpperCurvePoints = Array.ConvertAll(fadingCurvePoints, delegate(PointF p) {
          return new PointF(p.X + WaveFormBounds.Left, (1.0f - p.Y) * WaveFormBounds.Height * 0.5f + WaveFormBounds.Top);
        });
        var fadingLowerCurvePoints = Array.ConvertAll(fadingCurvePoints, delegate(PointF p) {
          return new PointF(p.X + WaveFormBounds.Left, (0.5f * p.Y + 0.5f) * WaveFormBounds.Height + WaveFormBounds.Top);
        });

        using (var penFadeEffect = new Pen(Color.FromArgb(0x80, 0x00, 0x40, 0xff))) {
          g.DrawCurve(penFadeEffect, fadingUpperCurvePoints);
          g.DrawCurve(penFadeEffect, fadingLowerCurvePoints);
        }

        // merges upper and lower curve points to fill
        var countPoints = fadingUpperCurvePoints.Length;

        Array.Resize(ref fadingUpperCurvePoints, fadingUpperCurvePoints.Length + fadingLowerCurvePoints.Length);
        Array.Reverse(fadingLowerCurvePoints);
        Array.Copy(fadingLowerCurvePoints, 0, fadingUpperCurvePoints, countPoints, fadingLowerCurvePoints.Length);

        g.FillPolygon(brushFadeEffect, fadingUpperCurvePoints, FillMode.Winding);
      } // using Brush, Pen

DrawFadeEffectDone:
      // wave form
      var truncatedWaveFormBounds = Rectangle.Truncate(WaveFormBounds);

      lock (cachedWaveFormImageLockObject) {
        if (cachedWaveFormImage == null)
          RedrawCachedWaveFormImage();

        if (cachedWaveFormImage != null)
          g.DrawImage(cachedWaveFormImage, truncatedWaveFormBounds, -(int)PixelsPerSecondOffset, 0, truncatedWaveFormBounds.Width, truncatedWaveFormBounds.Height, GraphicsUnit.Pixel);
      }

      // repeated point
      if (0 < track.TrackInfo.RepeatLength && 0.0f < WaveFormBounds.Width) {
        var prevSmoothingMode = g.SmoothingMode;

        g.SmoothingMode = SmoothingMode.None;

        using (var p = new Pen(Color.LightBlue)) {
          p.DashStyle = DashStyle.Dash;

          for (var repetition = track.TrackInfo.CalculateRepeatedTimes(TimeSpan.FromSeconds(hscrollOffset.Value)); ;repetition++) {
            var x = (float)(track.TrackInfo.CalculateRepeatPosition(repetition).TotalSeconds * PixelsPerSecondScale) + PixelsPerSecondOffset;

            if (x < 0.0f)
              continue;
            if (WaveFormBounds.Width <= x)
              break;

            x += WaveFormBounds.Left;

            g.DrawLine(p, x, WaveFormBounds.Top, x, WaveFormBounds.Bottom);
            g.DrawString((repetition + 1).ToString(), Control.DefaultFont, Brushes.White, x, WaveFormBounds.Top);
          }
        }

        g.SmoothingMode = prevSmoothingMode;
      }

      // anchors
      using (var p = new Pen(Color.White, 4.5f)) {
        p.SetLineCap(LineCap.NoAnchor, LineCap.RoundAnchor, DashCap.Flat);
        g.DrawLine(p, WaveFormBounds.Left, WaveFormBoundsMiddle, fadeStartX + 0.01f + WaveFormBounds.Left, WaveFormBoundsMiddle);
        g.DrawLine(p, fadeStartX + p.Width + WaveFormBounds.Left, WaveFormBoundsMiddle, fadeEndX + WaveFormBounds.Left, WaveFormBoundsMiddle);
        g.DrawLine(p, fadeEndX + p.Width + WaveFormBounds.Left, WaveFormBoundsMiddle, silenceEndX + WaveFormBounds.Left, WaveFormBoundsMiddle);
      }

DrawEffectDone:;
      // seek bar
      var seekBarX = (float)(seekBarPosition.TotalSeconds * PixelsPerSecondScale) + PixelsPerSecondOffset;

      if (0.0 <= seekBarX) {
        var prevSmoothingMode = g.SmoothingMode;
        g.SmoothingMode = SmoothingMode.None;
        using (var p = new Pen(Color.Red, 1)) {
          g.DrawLine(p, seekBarX + WaveFormBounds.Left, WaveFormBounds.Top, seekBarX + WaveFormBounds.Left, WaveFormBounds.Bottom);
        }
        g.SmoothingMode = prevSmoothingMode;
      }

      // track title
      if (track != null) {
        g.TextRenderingHint = TextRenderingHint.ClearTypeGridFit;

        using (var format = StringFormat.GenericDefault.Clone() as StringFormat) {
          format.Alignment = StringAlignment.Near;
          format.LineAlignment = StringAlignment.Far;
          format.FormatFlags |= StringFormatFlags.NoWrap;
          format.Trimming = StringTrimming.EllipsisCharacter;

          g.DrawString(track.TrackInfo.Title, Control.DefaultFont, Brushes.White, WaveFormBounds, format);
        }
      }

      base.OnPaint(e);
    }

    private void DiscardCachedWaveFormImage()
    {
      lock (cachedWaveFormImageLockObject) {
        if (cachedWaveFormImage != null) {
          cachedWaveFormImage.Dispose();
          cachedWaveFormImage = null;
        }
      }
    }

    private void RedrawCachedWaveFormImage()
    {
      var totalLength = track.Effect.TotalLength;
      var width = (int)(totalLength.TotalSeconds * PixelsPerSecondScale) + 1;
      var height = (int)WaveFormBounds.Height + 1;

      if (width <= 0 || height <= 0)
        return;

      cachedWaveFormImage = new Bitmap(width, height, PixelFormat.Format32bppArgb);

      using (var g = Graphics.FromImage(cachedWaveFormImage)) {
        g.Clear(Color.Transparent);

        if (waveForm == null)
          return;

#if true
        PointF[] waveFormLeft = null;
        PointF[] waveFormRight = null;

        var waveFormBoundsHalfHeight = WaveFormBounds.Height * 0.5f;
        var scale = WaveForm.WaveFormPartLengthInMilliseconds * PixelsPerSecondScale / 1000.0f;

        waveForm.Read(TimeSpan.Zero,
                      totalLength,
                      delegate(float[] samplesLeft, int start, int count) {
                        waveFormLeft = new PointF[count];

                        for (int offset = 0, index = start; offset < count; offset++, index++) {
                          waveFormLeft[offset] = new PointF(offset * scale, waveFormBoundsHalfHeight - samplesLeft[index] * waveFormBoundsHalfHeight);
                        }
                      },
                      delegate(float[] samplesRight, int start, int count) {
                        waveFormRight = new PointF[count];

                        for (int offset = 0, index = start; offset < count; offset++, index++) {
                          waveFormRight[offset] = new PointF(offset * scale, waveFormBoundsHalfHeight + samplesRight[index] * waveFormBoundsHalfHeight);
                        }
                      });

        if (waveFormLeft != null && 1 < waveFormLeft.Length &&
            waveFormRight != null && 1 < waveFormRight.Length) {
          using (var p = new Pen(Color.FromArgb(0xff, 0x40, 0x80, 0xff))) {
            g.DrawLines(p, waveFormLeft);
            g.DrawLines(p, waveFormRight);
          }
        }
#else
        float[] waveFormLeft;
        float[] waveFormRight;

        //waveForm.Get(TimeSpan.FromSeconds(hscrollOffset.Value), TimeSpan.FromSeconds(WaveFormBounds.Width + PixelsPerSecondScale), out waveFormLeft, out waveFormRight);
        waveForm.Get(TimeSpan.Zero, totalLength, out waveFormLeft, out waveFormRight);

        var waveFormBoundsHalfHeight = WaveFormBounds.Height * 0.5f;
        var scale = WaveForm.WaveFormPartLengthInMilliseconds * PixelsPerSecondScale / 1000.0f;

        if (1 < waveFormLeft.Length) {
          var index = 0;

          using (var p = new Pen(Color.FromArgb(0xff, 0x40, 0x80, 0xff))) {
            g.DrawLines(p, Array.ConvertAll(waveFormLeft, delegate(float level) {
              return new PointF(index++ * scale, waveFormBoundsHalfHeight - level * waveFormBoundsHalfHeight);
            }));
          }
        }

        if (1 < waveFormRight.Length) {
          var index = 0;

          using (var p = new Pen(Color.FromArgb(0xff, 0x40, 0x80, 0xff))) {
            g.DrawLines(p, Array.ConvertAll(waveFormRight, delegate(float level) {
              return new PointF(index++ * scale, waveFormBoundsHalfHeight + level * waveFormBoundsHalfHeight);
            }));
          }
        }
#endif
      } // using
    }

    protected override void OnClick(EventArgs e)
    {
      var point = PointToClient(Cursor.Position);

      if (WaveFormBounds.Contains(point)) {
        var newPosition = TimeSpan.FromSeconds(((point.X - WaveFormBounds.Left) - PixelsPerSecondOffset) / (double)PixelsPerSecondScale);

        if (track != null && newPosition <= track.Effect.TotalLength)
          OnSeekBarMoved(new SeekBarMovedEventArgs(newPosition));
      }

      base.OnClick(e);
    }

    protected virtual void OnSeekBarMoved(SeekBarMovedEventArgs e)
    {
      var ev = this.SeekBarMoved;

      if (ev != null)
        ev(this, e);

      if (!e.Cancel) {
        var lastSeekBarPosition = seekBarPosition;
        seekBarPosition = e.SeekBarPosition;
        UpdateSeekBarPosition(lastSeekBarPosition);
      }
    }

#region "EffectAnchor drag and drop operations"
    private EffectAnchor draggingEffectAnchor = null;
    private Point draggingEffectAnchorInitialLocation;
    private Point lastCursorPosition;

    private void EffectAnchor_MouseDown(object sender, MouseEventArgs e)
    {
      if (track == null)
        return;

      draggingEffectAnchor = sender as EffectAnchor;
      draggingEffectAnchorInitialLocation = draggingEffectAnchor.Location;
      lastCursorPosition = Cursor.Position;
    }

    private void EffectAnchor_MouseMove(object sender, MouseEventArgs e)
    {
      if (track == null)
        return;
      if (draggingEffectAnchor == null)
        return;

      var offset = (Cursor.Position.X - lastCursorPosition.X);
      var offsetTime = TimeSpan.FromSeconds(offset / (double)PixelsPerSecondScale);
      var effect = track.Effect;

      updatingEffectProperty = true;

      if (draggingEffectAnchor == fadingStartAnchor) {
        if (effect.FadingStart + offsetTime < TimeSpan.Zero)
          effect.FadingStart = TimeSpan.Zero;
        else
          effect.FadingStart += offsetTime;
      }
      else if (draggingEffectAnchor == fadingEndAnchor) {
        if (effect.FadingLength + offsetTime < TimeSpan.Zero)
          effect.FadingLength = TimeSpan.Zero;
        else
          effect.FadingLength += offsetTime;
      }
      else if (draggingEffectAnchor == trailingSilenceEndAnchor) {
        if (effect.TrailingSilenceLength + offsetTime < TimeSpan.Zero)
          effect.TrailingSilenceLength = TimeSpan.Zero;
        else
          effect.TrailingSilenceLength += offsetTime;
      }

      updatingEffectProperty = false;

      lastCursorPosition = Cursor.Position;

      UpdateAnchorLocation();

      Invalidate(Rectangle.FromLTRB(Math.Min(fadingStartAnchor.Left, draggingEffectAnchorInitialLocation.X) - 5, (int)WaveFormBounds.Top, (int)WaveFormBounds.Right, (int)WaveFormBounds.Bottom));
      Update();
    }

    private void EffectAnchor_MouseUp(object sender, MouseEventArgs e)
    {
      if (track == null)
        return;

      draggingEffectAnchor = null;

      UpdateScrollRange();

      Refresh();
    }
#endregion

    private void Effect_PropertyChanged(object sender, PropertyChangedEventArgs e)
    {
      if (sender == track.Effect) {
        if (e.PropertyName != "TrailingSilenceLength") {
          if (lastFadingStart < track.Effect.FadingStart)
            waveForm.Discard(lastFadingStart);
          else
            waveForm.Discard(track.Effect.FadingStart);
        }

        DiscardCachedWaveFormImage();

        lastFadingStart = track.Effect.FadingStart;

        if (!updatingEffectProperty)
          UpdateView();
      }
    }

    private void WaveForm_ReadWaveFormFinished(object sender, EventArgs e)
    {
      DiscardCachedWaveFormImage();

      InvalidateWaveFormBounds();
    }

    private void InvalidateWaveFormBounds()
    {
      if (InvokeRequired) {
        Invoke((MethodInvoker)InvalidateWaveFormBounds);
      }
      else {
        Invalidate(Rectangle.Ceiling(WaveFormBounds));
        Update();
      }
    }

    private Track track = null;
    private WaveForm waveForm = new WaveForm(null);
    private System.ComponentModel.Container components = new System.ComponentModel.Container();
    private HScrollBar hscrollOffset = null;
    private EffectAnchor fadingStartAnchor;
    private EffectAnchor fadingEndAnchor;
    private EffectAnchor trailingSilenceEndAnchor;
    private EffectAnchor[] anchors;
    private ToolTip toolTip;
    private TimeSpan seekBarPosition = TimeSpan.Zero;
    private bool updatingEffectProperty = false;
    private TimeSpan lastFadingStart = TimeSpan.Zero;
    private Bitmap cachedWaveFormImage;
    private object cachedWaveFormImageLockObject = new object();
  }
}
